
import inspect
import copy
from collections import OrderedDict

# from ginger.nav import Link
from djingles.utils import url_query_update

__all__ = ['Formatter', 'FormattedTable', 'FormattedObject']


class Formatter(object):

    __position = 1

    def __init__(self, label=None, attr=None, hidden=False, sortable=True, variants=None, reverse=False, empty=""):
        Formatter.__position += 1
        self.__position = Formatter.__position
        self.label = label
        self.hidden = hidden
        self.sortable = sortable
        self.reverse = reverse
        self.attr = attr
        self.empty = empty
        if variants is not None:
            if not isinstance(variants, (list, tuple)):
                variants = [variants]
            variants = set(variants)
        self.variants = variants

    def _update_position(self):
        Formatter.__position += 1
        self.__position = Formatter.__position

    def copy(self):
        instance = copy.copy(self)
        instance._update_position()
        return instance

    @property
    def position(self):
        return self.__position

    def format(self, value, name, source):
        return str(value)

    def extract(self, name, source, owner=None):
        if owner is not None:
            method = 'prepare_%s' % name
            func = getattr(owner, method, None)
            if func:
                return func(source)
        parts = name.split("__")
        result = source
        while parts:
            item = parts.pop(0)
            if isinstance(result, dict):
                result = result[item]
            else:
                result = getattr(result, item)
        return result

    def render(self, name, source, owner):
        value = self.extract(name, source, owner)
        return self.format(value, name, source)

    @classmethod
    def extract_from(cls, source, variant=None):
        result = sorted(inspect.getmembers(source, lambda a: isinstance(a, Formatter)),
            key=lambda p: p[1].position)
        result = [p for p in result if p[1].variants is None or variant in p[1].variants]
        return result


class FormattedValue(object):

    def __init__(self, name, prop, source, attrs=None, owner=None):
        self.name = name
        self.prop = prop
        self.source = source
        self.__attrs = attrs
        self.__owner = owner

    def get_absolute_url(self):
        try:
            func = self.__owner.get_cell_url
        except AttributeError:
            return None
        else:
            return func(self)

    @property
    def attrs(self):
        return self.__attrs(self) if self.__attrs else {}

    @property
    def label(self):
        label = self.prop.label
        return label if label is not None else self.name.replace("_", " ").title()

    @property
    def value(self):
        try:
            return self.prop.extract(self.name, self.source, self.__owner)
        except AttributeError as ex:
            raise ValueError("Error while accessing attribute %r in %r : %s" % (self.name, self.source, ex))

    def __getattr__(self, item):
        return getattr(self.prop, item)

    def __str__(self):
        value = self.value
        if value is None:
            return self.prop.empty
        result = self.prop.format(value, self.name, self.source)
        return str(result)


class FormattedObject(object):

    def __init__(self, obj, variant=None, **context):
        self.context = context
        self.variant = variant
        self.__prop_cache = OrderedDict((n,p) for (n,p) in Formatter.extract_from(self.__class__, variant=variant) if not p.hidden)
        self.source = obj
        data = self.data = OrderedDict()
        for name, prop in self.__prop_cache.items():
            data[name] = FormattedValue(name, prop, self.source, attrs=self.get_attrs, owner=self)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        try:
            return self.data[item]
        except KeyError:
            raise AttributeError(item)

    def __iter__(self):
        for name, prop in self.__prop_cache.items():
            yield self.data[name]

    def __len__(self):
        return len(self.__prop_cache)

    def __bool__(self):
        return len(self) > 0

    def get_attrs(self, value):
        return {}

    @classmethod
    def as_table(cls):
        props = dict(Formatter.extract_from(cls))
        for key, value in cls.__dict__.items():
            if not key.startswith("_") and callable(value):
                props[key] = value
        name = cls.__name__
        return type("%sTable"%name, (FormattedTable,), props)


class FormattedTableColumn(object):

    __inited = False

    def __init__(self, name, prop, table):
        self.name = name
        self.prop = prop
        self.table = table
        self.__inited = True

    @property
    def label(self):
        return self.prop.label or self.name.capitalize()

    def __setattr__(self, key, value):
        if self.__inited and key not in __dict__:
            setattr(self.prop, key, value)
        else:
            self.__dict__[key] = value

    def __getattr__(self, item):
        return getattr(self.prop, item)

    def __getitem__(self, item):
        return getattr(self.prop, item)

    def values(self):
        for obj in self:
            yield obj.value

    def __iter__(self):
        for obj in self.table.source:
            yield FormattedValue(self.name, self.prop, obj, self.table.get_cell_attrs)

    def __len__(self):
        return len(self.table.source)

    def __bool__(self):
        return len(self) > 0


class FormattedTableColumnSet(object):

    def __init__(self, table, values):
        self.columns = OrderedDict((n, FormattedTableColumn(n, p, table)) for n,p in values)

    def visible_columns(self):
        return [col for col in self.columns.values() if not col.hidden]

    def hidden_columns(self):
        return [col for col in self.columns.values() if col.hidden]

    def keys(self):
        return self.columns.keys()

    def __iter__(self):
        for value in self.columns.values():
            yield value

    def __contains__(self, item):
        return item in self.columns

    def __getitem__(self, item):
        return self.columns[item]

    def __getattr__(self, item):
        return self.columns[item]

    def __len__(self):
        return len(self.columns)

    def __bool__(self):
        return len(self) > 0


class FormattedTableRow(object):

    def __init__(self, index, source, table, kind=None):
        self.index = index
        self.source = source
        self.table = table
        self.kind = kind
        self.data = OrderedDict()
        for column in self.table.columns:
            self.data[column.name] = FormattedValue(column.name, column.prop, source,
                                                    attrs=self.table.get_cell_attrs,
                                                    owner=table)

    def __getitem__(self, item):
        return self.data[item]

    @property
    def object(self):
        return self.source

    @property
    def attrs(self):
        return self.table.get_row_attrs(self)

    def __iter__(self):
        for column in self.table.columns:
            yield self.data[column.name]

    def __len__(self):
        return len(self.table.columns)

    def __bool__(self):
        return len(self) > 0


class FooterRow:

    __inited = False

    def __init__(self, table, label):
        self.data = {}
        for column in table.columns:
            self.data[column.name] = None
        self.label = label
        self.__inited = True

    def __getattr__(self, item):
        return self.data[item] if item in self.data else None

    def __setattr__(self, key, value):
        if self.__inited:
            self.data[key] = value
        else:
            self.__dict__[key] = value

    def __iter__(self):
        return iter(self.data)

    def to_row(self):
        return self.data


class FormattedTable(object):

    def __init__(self, source, sort_key=None, sort_field=None, variant=None, **context):
        self.context = context
        self.variant = variant
        self.columns = FormattedTableColumnSet(self, Formatter.extract_from(self.__class__, variant=variant))
        self.source = source
        self.sort_field = sort_field
        self.sort_key = sort_key

    def build_links(self, request):
        data = request.GET
        for col in self.columns.visible_columns():
            if self.sort_key and self.sort_field:
                field = self.sort_field
                code = field.get_value_for_name(col.name)
                value = data.get(self.sort_key, "")
                reverse = value.startswith("-")
                if reverse:
                    value = value[1:]
                is_active = code == value
                next_value = "-%s" % code if not reverse and is_active else code
                mods = {self.sort_key: next_value}
            else:
                is_active = False
                reverse = False
                mods = {}
            url = url_query_update(request.build_absolute_uri(), mods) if mods else None
            link = Link(content=col.label, url=url, is_active=is_active, reverse=reverse, sortable=col.sortable, column=col)
            yield link

    @property
    def object_list(self):
        return self.source

    def footer_row(self, label):
        row = FooterRow(self, label)
        return row

    @property
    def footer(self):
        index = 0
        for obj in self.get_footer_rows():
            yield FormattedTableRow(index, obj, self)
            index += 1

    def __iter__(self):
        index = 0
        for obj in self.source:
            yield FormattedTableRow(index, obj, self)
            index += 1

    def __len__(self):
        return len(self.source)

    def __bool__(self):
        return len(self) > 0

    def get_row_attrs(self, row):
        return {}

    def get_cell_attrs(self, cell):
        return {}

    def get_footer_rows(self):
        return ()




from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import reverse
from jinja2 import Environment

from .library import jinja2_filter, jinja2_function, library


__all__ = ['jinja2_filter', 'jinja2_function', 'environment']


def inject_filters(env):
    from django.template.defaultfilters import register
    for name, func in register.filters.items():
        if name not in env.filters:
            env.filters[name] = func


def inject_functions(env):
    pass


def inject_library(env):
    for key, func in library.functions.items():
        env.globals[key] = func

    for key, func in library.filters.items():
        env.filters[key] = func


def environment(**options):
    env = Environment(**options)
    env.globals.update({
        'static': staticfiles_storage.url,
        'url': reverse,
    })
    inject_filters(env)
    inject_functions(env)
    inject_library(env)
    return env





import inspect
from collections import OrderedDict

from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from django import forms
from django.utils.encoding import force_text
from djingles.formatters import Formatter
from djingles.forms.fields import GingerSortField
from djingles import utils


__all__ = ['ActionFormMixin', 'ActionModelForm', 'ActionForm', 'TableSortField', 'FilterForm',
           'FilterModelForm', 'FilterFormMixin', 'action_model_factory']


class ActionFormMixin(object):

    confirmation_message = None
    failure_message = None
    success_message = None

    def __init__(self, **kwargs):
        self.optional_fields = kwargs.pop("optional_fields", set())
        parent_cls = forms.Form if not isinstance(self, forms.ModelForm) else forms.ModelForm
        constructor = parent_cls.__init__
        parent_constructor = super(ActionFormMixin, self).__init__
        keywords = set(inspect.getargspec(constructor).args)
        func = lambda a: getattr(a, 'im_func', a)
        parent_keywords = set(inspect.getargspec(parent_constructor).args) \
            if func(parent_constructor) is not func(constructor) else set()
        context = {}
        for key in kwargs.copy():
            if key in keywords and key not in parent_keywords:
                continue
            value = kwargs.pop(key) if key not in parent_keywords else kwargs.get(key)
            context[key] = value
        super(ActionFormMixin, self).__init__(**kwargs)
        self.context = context
        prepare = getattr(self, 'prepare_%s' % self.get_action_name(), None)
        if prepare is not None:
            prepare()

    def get_action_name(self):
        return self.context.get("action", "execute")

    def get_action_method(self):
        func = self.get_action_name() or 'execute'
        return getattr(self, func)

    def __bind_fields(self):
        key = "__bound_with_form"
        optionals = set(self.optional_fields if self.optional_fields is not None else self.fields.keys())
        for name, field in self.fields.items():
            if name in optionals:
                field.required = False
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    def process_context(self, func=None):
        context = self.context.copy()

        if hasattr(self, "cleaned_data") and hasattr(self, "save"):
            instance = self.save(commit=False)
            context["instance"] = instance

        context["data"] = self.cleaned_data

        spec = inspect.getargspec(func or self.get_action_method())
        if spec.varargs:
            raise ImproperlyConfigured("Form action cannot have variable arguments")
        if spec.keywords:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}

    def process_result(self, result):
        return result

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        if self.success_message:
            return self.success_message.format(**self.context)
        return self.success_message

    def get_failure_message(self):
        if self.failure_message:
            return self.failure_message.format(**self.context)
        return self.failure_message

    def get_confirmation_message(self):
        if self.confirmation_message:
            return self.confirmation_message.format(**self.context)
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls))

    @classmethod
    def call(cls, *kwargs):
        return cls(**kwargs).run()

    def run(self):
        if not self.is_valid():
            return False, self.errors
        else:
            return True, self.result

    def full_clean(self):
        self.__bind_fields()
        super(ActionFormMixin, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            result = None
            try:
                if self.is_bound and not self._errors:
                    context = self.process_context()
                    result = self.get_action_method()(**context)
                    result = self.process_result(result)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def select_fields(self, exclude=None, include=None, **extras):
        if exclude is not None:
            for f in exclude:
                self.fields.pop(f, None)
        if include is not None:
            include = set(include)
            for f in list(self.fields.keys()):
                if f not in include:
                    self.fields.pop(f)
        if extras:
            for f in extras:
                self.fields[f] = extras[f]

    def execute(self, **kwargs):
        return {}


class ActionForm(ActionFormMixin, forms.Form):
    pass


class ActionModelForm(ActionFormMixin, forms.ModelForm):

    confirmation_message = "Please confirm that you really wish to {action} this object ?"

    def prepare_delete(self):
        self.select_fields(include=())

    def delete(self):
        self.instance.delete()

    def update(self):
        return self.save()

    def create(self):
        return self.save()


class FilterFormMixin(ActionFormMixin):

    def __init__(self, queryset, **kwargs):
        self.queryset = queryset
        kwargs.setdefault("optional_fields", None)
        super(FilterFormMixin, self).__init__(**kwargs)

    def perform_action(self):
        queryset = self.perform_filter()
        action = self.get_action_method()
        return action(queryset)

    def perform_filter(self):
        queryset = self.queryset
        if not self.is_valid():
            return self.invalid_queryset(queryset)
        data = self.cleaned_data
        allowed = set(self.fields.keys())
        if hasattr(self, "before_filters"):
            queryset = self.before_filters(queryset, data)
        for name, value in six.iteritems(data):
            if name not in allowed or value in (None, '', [], ()):
                continue
            kwargs = {}
            field = self.fields[name]
            if hasattr(self, "filter_%s" % name):
                result = getattr(self, "filter_%s" % name)(queryset, value, data)
            elif hasattr(field, "filter_queryset"):
                result = field.filter_queryset(queryset, value, self[name])
            else:
                if isinstance(value, (tuple,list)):
                    name = '%s__in' % name
                kwargs[name] = value
                result = queryset.filter(**kwargs)
            if result is not None:
                queryset = result
        if hasattr(self, "after_filters"):
            queryset = self.after_filters(queryset, data)
        return queryset

    def invalid_queryset(self, queryset):
        return queryset.none()


class FilterForm(FilterFormMixin, forms.Form):
    pass


class FilterModelForm(FilterFormMixin, forms.ModelForm):
    pass


class TableSortField(GingerSortField):

    def __init__(self, table_class, **kwargs):
        column_dict = OrderedDict(Formatter.extract_from(table_class))
        self.reverse = kwargs.pop("reverse", False)
        choices = [(name, col.label or name.title()) for name, col in six.iteritems(column_dict) if not col.hidden]
        super(TableSortField, self).__init__(choices=choices, **kwargs)
        self.table_class = table_class

    def handle_queryset(self, queryset, value, bound_field):
        text_value = force_text(value) if value is not None else None
        if not text_value:
            return queryset
        reverse = text_value.startswith("-")
        column_dict = self.table_class.get_column_dict()
        name = text_value[1:] if reverse else text_value
        name = self.field_map[name]
        col = column_dict[name]
        if not col.sortable:
            return queryset
        attr = col.attr or name
        if col.reverse:
            reverse = not reverse
        if reverse:
            attr = "-%s" % attr
        return queryset.order_by(attr)


def action_model_factory(model_class, include=None, exclude=(), **kwargs):
    name = "%sActionForm" % model_class.__name__
    meta = type("Meta", (), {"include": include, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (ActionModelForm, ), kwargs)



import random
from ginger.formatters.base import *


class SimpleFormatter(Formatter):

    def format(self, value, name, source):
        return ">>>>>>> %s" % value


class Item:

    def __init__(self):
        self.a = random.randint(0, 10)
        self.b = random.randint(0, 10)
        self.c = random.randint(0, 10)

source = [Item() for i in range(100)]


class TestPropertySet(FormattedObject):
    a = SimpleFormatter()
    b = Formatter()
    c = Formatter()

    def get_row_attrs(self, row):
        return {"class": "class%s" % row.index}

    def get_cell_attrs(self, cell):
        return {"name": "asdas", "class": "something"}


TestPropertyTable = TestPropertySet.as_table()


if __name__ == '__main__':
    prop = TestPropertyTable(source)
    for p in prop:
        print([a.name for a in p], [a.value for a in p])





import datetime
import inspect
from django.core.exceptions import ImproperlyConfigured
from django.http.request import QueryDict
from django.utils import six
from django import forms
from django.core.paginator import Page

from ginger.exceptions import ValidationFailure
from ginger import utils, paginator
from ginger import html
import warnings


__all__ = ['GingerModelForm', 
           'GingerForm',
           'GingerSearchModelForm',
           'GingerSearchForm',
           'GingerSafeEmptyTuple',
           'GingerFormMixin',
           'GingerSearchFormMixin',
           'GingerDataForm',
           'GingerDataModelForm',
           'GingerModelFormSet']



class GingerSafeEmptyTuple(tuple):
    def __len__(self):
        return 1

class GingerFormMixinBase(object):
    failure_message = None
    success_message = None
    confirmation_message = None

    def __init__(self, **kwargs):
        parent_cls = forms.Form if not isinstance(self, forms.ModelForm) else forms.ModelForm
        constructor = parent_cls.__init__
        keywords = set(inspect.getargspec(constructor).args)
        context = {}
        for key in kwargs.copy():
            if key in keywords:
                continue
            value = kwargs.pop(key)
            context[key] = value
        super(GingerFormMixinBase, self).__init__(**kwargs)
        self.context = context

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        return self.success_message

    def get_failure_message(self):
        return self.failure_message

    def get_confirmation_message(self):
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls).encode('utf-8'))

    def process_context(self):
        context = self.context.copy()
        if hasattr(self, "cleaned_data"):
            if hasattr(self, "save"):
                instance = self.save(commit=False)
                context["instance"] = instance
            context["data"] = self.cleaned_data
        spec = inspect.getargspec(self.execute)
        if spec.varargs:
            raise ImproperlyConfigured("Form.execute cannot have variable arguments")
        if spec.keywords:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}


    def full_clean(self):
        super(GingerFormMixinBase, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            result = None
            try:
                if self.is_bound :
                    context = self.process_context()
                    result = self.execute(**context)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def execute(self, **kwargs):
        return {}

    @classmethod
    def is_submitted(cls, data):
        return data and (any(k in data for k in cls.base_fields) or cls.submit_name() in data)

    @classmethod
    def submit_name(cls):
        return "submit-%s" % cls.class_oid()


class GingerFormMixin(object):

    failure_message = None
    success_message = None
    confirmation_message = None
    ignore_errors = False
    use_defaults = False
    template_name = None

    def __init__(self, **kwargs):
        parent_cls = forms.Form if not isinstance(self, forms.ModelForm) else forms.ModelForm
        constructor = parent_cls.__init__
        parent_constructor = super(GingerFormMixin, self).__init__
        keywords = set(inspect.getargspec(constructor).args)
        func = lambda a: getattr(a, 'im_func', a)
        parent_keywords = set(inspect.getargspec(parent_constructor).args) \
            if func(parent_constructor) is not func(constructor) else set()
        self.use_defaults = kwargs.pop("use_defaults", self.use_defaults)
        if "ignore_errors" in kwargs:
            self.ignore_errors = kwargs.pop("ignore_errors")
        context = {}
        for key in kwargs.copy():
            if key in keywords and key not in parent_keywords:
                continue
            value = kwargs.pop(key) if key not in parent_keywords else kwargs.get(key)
            context[key] = value
        super(GingerFormMixin, self).__init__(**kwargs)
        self.context = context
        self.merge_defaults()

    def get_template_names(self):
        template_name = self.template_name
        if not template_name:
            raise ImproperlyConfigured("Not template defined for this form")
        return [template_name]

    def insert_null(self, field_name, label, initial=""):
        field = self.fields[field_name]
        if initial is None:
            initial = field.empty_value
        field.required = False
        field.initial = initial
        choices = list(field.choices)
        top = choices[0][0]
        if top == field.empty_value or not top:
            choices = choices[1:]
        choices.insert(0, (initial, label))
        field.choices = tuple(choices)

    def _bind_fields(self):
        key = "__bound_with_form"
        for field in self.fields.values():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    def field_range(self, first, last=None, step=None):
        return html.field_range(self, first, last, step, hidden=True)

    def iter_fields(self, names):
        return (self[field] for field in names)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.field_range(item.start, item.stop, item.step)
        return super(GingerFormMixin, self).__getitem__(item)

    def merge_defaults(self):
        if self.use_defaults:
            data = QueryDict('', mutable=True)
            if self.data:
                data.update(self.data)
            initial = self.initial_data
            for key in initial:
                if key in data:
                    continue
                value = initial[key]
                name = self.add_prefix(key)
                if value is not None:
                    if isinstance(value, (list, tuple)):
                        data.setlistdefault(name, value)
                    else:
                        data.setdefault(name, value)
            self.data = data

    def process_context(self):
        context = self.context.copy()
        if not isinstance(self, GingerSearchFormMixin) and hasattr(self, "cleaned_data"):
            if hasattr(self, "save"):
                instance = self.save(commit=False)
                context["instance"] = instance
            context["data"] = self.cleaned_data
        spec = inspect.getargspec(self.execute)
        if spec.varargs:
            raise ImproperlyConfigured("Form.execute cannot have variable arguments")
        if spec.keywords:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}

    @property
    def initial_data(self):
        fields = self.fields
        result = {}
        for name,field in six.iteritems(fields):
            data = self.initial.get(name, field.initial)
            if callable(data):
                data = data()
                if (isinstance(data, (datetime.datetime, datetime.time)) and
                        not getattr(field.widget, 'supports_microseconds', True)):
                    data = data.replace(microsecond=0)
            if data and isinstance(field, forms.MultiValueField):
                for i, f in enumerate(field.fields):
                    key = "%s_%s" % (name, i)
                    result[key] = data[i]
            elif data is not None:
                result[name] = data
        return result

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        return self.success_message

    def get_failure_message(self):
        return self.failure_message

    def get_confirmation_message(self):
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls))

    @classmethod
    def is_submitted(cls, data):
        return data and (any(k in data for k in cls.base_fields) or cls.submit_name() in data)

    @classmethod
    def submit_name(cls):
        return "submit-%s" % cls.class_oid()

    def run(self):
        if not self.is_valid() and not self.ignore_errors:
            raise ValidationFailure(self)
        return self.result

    def full_clean(self):
        super(GingerFormMixin, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            result = None
            try:
                self._bind_fields()
                if self.is_bound and (not self._errors or self.ignore_errors):
                    context = self.process_context()
                    result = self.execute(**context)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def execute(self, **kwargs):
        return {}

    def as_html(self):
        return html

    def to_json(self):
        return {
            'message': self.get_success_message(),
            'data': self.run()
        }


class GingerModelForm(GingerFormMixin, forms.ModelForm):
    pass

class GingerForm(GingerFormMixin, forms.Form):
    pass


class GingerSearchFormMixin(GingerFormMixin):

    per_page = 20
    page_limit = 10
    parameter_name = "page"
    ignore_errors = True
    use_defaults = True

    def _post_clean(self):
        """
            This override is needed so as to avoid modelform validation during clean
        """

    def get_sort_field(self):
        from .fields import GingerSortField
        try:
            return next(self[name] for name, f in six.iteritems(self.fields) if isinstance(f, GingerSortField))
        except StopIteration:
            return None

    def get_queryset(self, **kwargs):
        queryset = self.context.get("queryset")
        if queryset is not None:
            return queryset
        return self.queryset

    def execute(self, **kwargs):
        return self.process_queryset_filters(**kwargs)

    def get_queryset_filter_names(self):
        return self.fields.keys()

    def process_queryset_filters(self, page=None, parameter_name="page",
                      page_limit=10, per_page=20, **kwargs):
        queryset = self.get_queryset(**kwargs)
        data = self.cleaned_data if self.is_bound else self.initial_data
        allowed = set(self.get_queryset_filter_names())
        if hasattr(self, "before_filters"):
            queryset = self.before_filters(queryset, data)
        for name, value in six.iteritems(data):
            if name not in allowed or value in (None, '', [], ()):
                continue
            kwargs = {}
            field = self.fields[name]
            if hasattr(self, "filter_%s" % name):
                result = getattr(self, "filter_%s" % name)(queryset, value, data)
            elif hasattr(self, "handle_%s" % name):
                result = getattr(self,"handle_%s" % name)(queryset, value, data)
            elif hasattr(field, "handle_queryset"):
                result = field.handle_queryset(queryset, value, self[name])
            else:
                if isinstance(value, (tuple,list)):
                    name = '%s__in' % name
                kwargs[name] = value
                result = queryset.filter(**kwargs)
            if result is not None:
                queryset = result
        if hasattr(self, "after_filters"):
            queryset = self.after_filters(queryset, data)
        if page is not None:
            queryset = self.paginate(queryset, page,
                                 parameter_name=parameter_name,
                                 page_limit=page_limit, per_page=per_page)
        return queryset

    @staticmethod
    def paginate(object_list, page, **kwargs):
        return paginator.paginate(object_list, page, **kwargs)

    def is_paginated(self):
        return 'page' in self.context

    def to_json(self):
        warnings.warn("Form.to_json is deprecated and shall be removed in ginger 1.0", DeprecationWarning)
        result = self.run()
        if isinstance(result, Page):
            return {
                'data': result.object_list,
                'page': result
            }
        else:
            return {
                'data': result
            }


class GingerSearchModelForm(GingerSearchFormMixin, forms.ModelForm):
    pass


class GingerSearchForm(GingerSearchFormMixin, forms.Form):
    pass


class GingerDataFormMixin(GingerSearchFormMixin):

    def execute(self, **kwargs):
        result = super(GingerDataFormMixin, self).execute(**kwargs)
        schema_cls = self.get_dataset_class()
        dataset = schema_cls()
        self.load_dataset(dataset, result)
        self.process_dataset_filters(dataset)
        return dataset

    def load_dataset(self, dataset, data_source):
        dataset.extend(data_source)

    def process_dataset_filters(self, dataset, **kwargs):
        cleaned_data = self.cleaned_data if self.is_bound else self.initial_data
        for name in self.get_dataset_filter_names():
            value = cleaned_data.get(name)
            field = self.fields.get(name)
            if value is None or value == "":
                continue
            if hasattr(self, "handle_%s" % name):
                getattr(self,"handle_%s" % name)(dataset, value, cleaned_data)
            elif field and hasattr(field, "handle_dataset"):
                field.handle_dataset(dataset, value, self[name])
        sort_field = self.get_sort_field()
        if sort_field:
            dataset.sort_parameter_name = sort_field.html_name
            dataset.sort_field = sort_field.field
        return dataset

    def get_dataset_class(self):
        try:
            return next(f.dataset_class for f in six.itervalues(self.fields)
                        if hasattr(f, "dataset_class"))
        except StopIteration:
            return self.dataset_class

    def get_dataset_filter_names(self):
        names = set(name for name, f in six.iteritems(self.fields)
                    if getattr(f, "process_list", False))
        names.update(getattr(self, "dataset_filters", ()))
        return names

    def get_queryset_filter_names(self):
        names = super(GingerDataFormMixin, self).get_queryset_filter_names()
        dataset_filters = set(self.get_dataset_filter_names())
        return [name for name in names if name not in dataset_filters]

    def to_json(self):
        result = self.run()
        return {
            "data": result.rows,
            "aggregates": result.aggregates.rows,
            "page": result.object_list if result.is_paginated() else None,
            "schema": [col.to_json() for col in result.columns]
        }


class GingerDataModelForm(GingerDataFormMixin, forms.ModelForm):
    pass


class GingerDataForm(GingerDataFormMixin, forms.Form):
    pass


class GingerModelFormSet(forms.BaseModelFormSet):
    failure_message = None
    success_message = None
    confirmation_message = None

    def __init__(self, **kwargs):
        parent_cls = forms.BaseFormSet if not isinstance(self, forms.BaseModelFormSet) else forms.BaseModelFormSet
        constructor = parent_cls.__init__
        keywords = set(inspect.getargspec(constructor).args)
        context = {}
        for key in kwargs.copy():
            if key in keywords:
                continue
            value = kwargs.pop(key)
            context[key] = value
        super(GingerModelFormSet, self).__init__(**kwargs)
        self.context = context

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        return self.success_message

    def get_failure_message(self):
        return self.failure_message

    def get_confirmation_message(self):
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls).encode('utf-8'))

    def process_context(self):
        context = self.context.copy()
        if hasattr(self, "cleaned_data"):
            if hasattr(self, "save"):
                instance = self.save(commit=False)
                context["instance"] = instance
            context["data"] = self.cleaned_data
        spec = inspect.getargspec(self.execute)
        if spec.varargs:
            raise ImproperlyConfigured("Form.execute cannot have variable arguments")
        if spec.keywords:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}


    def full_clean(self):
        super(GingerModelFormSet, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            result = None
            try:
                if self.is_bound and not any(self._errors) and not self._non_form_errors:
                    context = self.process_context()
                    result = self.execute(**context)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def execute(self, **kwargs):
        return {}

    @classmethod
    def is_submitted(cls, data):
        return data and (any(k in data for k in cls.base_fields) or cls.submit_name() in data)

    @classmethod
    def submit_name(cls):
        return "submit-%s" % cls.class_oid()

    def add_error(self, name, errors):
        if name is None:
            self._non_form_errors.append(errors)



from django.conf import settings
import importlib
from django.utils.encoding import force_text
from . import common as html, forms

engines = None


__all__ = ['get_engine', 'render_form', 'render_column',
           'render_container', 'render_row', 'render_form_end', 'render_form_field',
           'render_form_footer', 'render_form_header', 'render_form_start', 'row', 'container', 'col', 'field']


def get_engine(name="default"):
    global engines
    if engines is None:
        engines = {}
    if name not in engines:
        config = getattr(settings, "GINGER_FRONTEND_ENGINE", {"default": "ginger.html.layouts.FrontendEngine"})
        module_name, class_name = config[name].rsplit(".", 1)
        module = importlib.import_module(module_name)
        engine = getattr(module, class_name)()
        engines[name] = engine
    return engines[name]


def render_form_field(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form_field(*args, **kwargs)


def render_form(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form(*args, **kwargs)


def render_form_start(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form_start(*args, **kwargs)


def render_form_end(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form_end(*args, **kwargs)


def render_form_header(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form_header(*args, **kwargs)


def render_form_footer(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.form_footer(*args, **kwargs)


def render_row(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.row(*args, **kwargs)


def render_column(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.column(*args, **kwargs)


def render_container(name, *args, **kwargs):
    engine = get_engine(name)
    return engine.container(*args, **kwargs)



class FrontendEngine(object):

    def form_hidden_field(self, context, field):
        return force_text(field)

    def form_csrf(self, context):
        request = context['request']
        return forms.form_csrf_tag(request)

    def form_start(self, context, form, **attrs):
        is_csrf = attrs.get("method", "post").lower() == "post"
        attrs = forms.form_attrs(form, **attrs)
        form_tag = "<form {}>".format(attrs)
        mgmt = force_text(getattr(form, "management_form", ""))
        csrf = self.form_csrf(context) if is_csrf else ""
        return "%s%s%s" % (form_tag, csrf, mgmt)

    def form_end(self, context, form , **attrs):
        return "</form>"

    def form_visible_field(self, context, field):
        return force_text(field)

    def form_header(self, context, form):
        return ""

    def form_footer(self, context, form):
        return ""

    def form_field(self, context, tag):
        field = tag.field
        return self.form_visible_field(context, field) if field.is_hidden \
            else self.form_hidden_field(context, field)

    def row(self, context, tag):
        return tag.mutate(html.div(class_="row"))

    def column(self, context, tag):
        return tag.mutate(html.div(class_="col-md-2"))

    def container(self, context, tag):
        return tag.mutate(html.div(class_="container"))

    def form(self, context, form, **attrs):
        start = self.form_start(context, form, **attrs)
        end = self.form_end(context, form, **attrs)
        if hasattr(form, "render"):
            content = html.empty[form.render(context, **attrs)]
            fields = force_text(content.render(context, **attrs))
        else:
            fields = " ".join([self.form_field(context, f) for f in form])
        header = self.form_header(context, form)
        footer = self.form_footer(context, form)
        errors = self.form_non_field_errors(context, form)
        return "{start}{errors}{header}{fields}{footer}{end}".format(**locals())

    def form_non_field_errors(self, context, form):
        return force_text(form.non_field_errors())


class Container(html.Element):
    def render(self, context, *args, **kwargs):
        engine = get_engine()
        return str(engine.container(context, self, *args, **kwargs))


class Column(html.Element):
    def render(self, context, *args, **kwargs):
        engine = get_engine()
        return str(engine.column(context, self, *args, **kwargs))


class Row(html.Element):
    def render(self, context, *args, **kwargs):
        engine = get_engine()
        return str(engine.row(context, self, *args, **kwargs))


class Field(html.Element):

    def __init__(self, *args, **kwargs):
        super(Field, self).__init__("div")

    def __call__(self, field, **kwargs):
        cp = super(Field, self).__call__(**kwargs)
        cp.field = field
        return cp

    def render(self, ctx, *args, **kwargs):
        engine = get_engine()
        return str(engine.form_field(ctx, self, *args, **kwargs))


row = Row("div")
field = Field()
col = Column("div")
container = Container("div")



import re
import warnings
import mimetypes
# import urllib2
from django.utils.encoding import force_text
from django.utils import six
import os
from urllib.request import urlopen
from urllib.parse import urlparse
from django import forms
from django.core.validators import URLValidator
from django.core.files.uploadedfile import SimpleUploadedFile

from djingles import utils
from djingles.utils import feet_inches_to_cm, cm_to_feet_inches


__all__ = ["FileOrUrlInput", "HeightField", "HeightWidget", "SortField", "GingerDataSetField", "GingerSortField", 
           "GingerPageField"]


class FileOrUrlInput(forms.ClearableFileInput):
    
    def download_url(self, name, url):
        validate = URLValidator()
        try:
            validate(url)
        except forms.ValidationError as _:
            raise
            return None
        
        parsed_url = urlparse(url)
        path = parsed_url[2].strip("/")
        name = os.path.basename(path)
        opener = urllib2.build_opener()
        opener.addheaders = [('User-agent', 'Mozilla/5.0')]
        ze_file = opener.open(url).read()
        file_obj = SimpleUploadedFile(name=name, content=ze_file, content_type=mimetypes.guess_type(name))
        file_obj.url = url
        return file_obj

    def value_from_datadict(self, data, files, name):
        if name in data and name not in files:
            url = forms.HiddenInput().value_from_datadict(data, files, name)
            result = self.download_url(name, url) if url and isinstance(url, six.string_types) else None
            files = files.copy() if files else {}
            files[name] = result
        return super(FileOrUrlInput, self).value_from_datadict(data, files, name)


class HeightWidget(forms.MultiWidget):

    def __init__(self, *args, **kwargs):
        widgets = [forms.TextInput(attrs={'placeholder': '5', 'size': '3'}), forms.TextInput(attrs={'placeholder': '6',
                                                                                                    'size': '3'})]
        super(HeightWidget,self).__init__(widgets, *args, **kwargs)

    def decompress(self, value):
        if value:
            result = cm_to_feet_inches(value)
            return result
        else:
            return [None,None]

    def format_output(self, rendered_widgets):
        return "%s ft   %s inches" % tuple(rendered_widgets)


class HeightField(forms.MultiValueField):
    widget = HeightWidget

    def __init__(self, *args, **kwargs):
        kwargs.pop('min_value',None)
        errors = self.default_error_messages.copy()
        if 'error_messages' in kwargs:
            errors.update(kwargs['error_messages'])
        reqd = kwargs.setdefault('required', False)
        fields = (
            forms.IntegerField(min_value=0,required=reqd),
            forms.IntegerField(min_value=0,required=reqd),
        )
        super(HeightField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        if data_list and all(d is not None for d in data_list):
            feet, inches = data_list
            return feet_inches_to_cm(feet, inches)
        return None


class GingerSortField(forms.ChoiceField):

    def __init__(self, choices=(), toggle=True, **kwargs):
        kwargs.setdefault("required", False)
        kwargs.setdefault("widget", forms.HiddenInput)
        super(GingerSortField, self).__init__(choices=choices, **kwargs)
        self.toggle = toggle
        field_map = {}
        new_choices = []
        for i, (value, label) in enumerate(choices):
            position = str(i)
            new_choices.append((position, label))
            field_map[position] = re.sub(r'\s+', ' ', value.strip())
        self.choices = tuple(new_choices)
        self.field_map = field_map

    def valid_value(self, value):
        "Check to see if the provided value is a valid choice"
        text_value = force_text(value)
        if text_value.startswith("-"):
            text_value = text_value[1:]
        return text_value in self.field_map or super(GingerSortField, self).valid_value(text_value)

    def build_links(self, request, bound_field):
        value = bound_field.value()
        field_name = bound_field.name
        text_value = force_text(value) if value is not None else None
        for k, v in self.choices:
            content = force_text(v)
            key = force_text(k)
            is_active = text_value and text_value == key
            if is_active and self.toggle:
                next_value = key if text_value.startswith("-") else "-%s" % key
            else:
                next_value = key
            url = utils.get_url_with_modified_params(request, {field_name: next_value})
            yield ui.Link(url=url, content=content, is_active=is_active)

    def invert_sign(self, name, neg):
        if name.startswith("-"):
            neg = not neg
        return "%s%s" % ("-" if neg else "", name.lstrip("-"))

    def handle_queryset(self, queryset, key, bound_field):
        neg = key.startswith("-")
        value = self.field_map[key.lstrip("-")]
        invert = lambda a: self.invert_sign(a, neg)
        values = map(invert, value.split())
        return queryset.order_by(*values)

    def get_value_for_name(self, name):
        for value, key in six.iteritems(self.field_map):
            if name == key:
                return value


class GingerDataSetField(GingerSortField):

    def __init__(self, dataset_class, process_list=False, **kwargs):
        column_dict = dataset_class.get_column_dict()
        self.reverse = kwargs.pop("reverse", False)
        choices = [(name, col.label or name.title()) for name, col in six.iteritems(column_dict) if not col.hidden]
        super(GingerDataSetField, self).__init__(choices=choices, **kwargs)
        self.dataset_class = dataset_class
        self.process_list = process_list

    def handle_queryset(self, queryset, value, bound_field):
        text_value = force_text(value) if value is not None else None
        if not text_value:
            return queryset
        reverse = text_value.startswith("-")
        column_dict = self.dataset_class.get_column_dict()
        name = text_value[1:] if reverse else text_value
        name = self.field_map[name]
        col = column_dict[name]
        if not col.sortable:
            return queryset
        attr = col.attr or name
        if col.reverse:
            reverse = not reverse
        if reverse:
            attr = "-%s" % attr
        return queryset.order_by(attr)

    def handle_dataset(self, dataset, value, bound_field):
        text_value = force_text(value) if value is not None else None
        if not text_value:
            return
        reverse = text_value.startswith("-")
        value = text_value[1:] if reverse else text_value
        name = self.field_map[value]
        column = dataset.columns[name]
        if column.reverse:
            reverse = not reverse
        column.sort(reverse=reverse)


class SortField(GingerSortField):

    def __init__(self, *args, **kwargs):
        super(SortField, self).__init__(*args, **kwargs)
        warnings.warn("Please use GingerSortField instead of SortField", DeprecationWarning)





import functools
import jinja2


class _Library:

    def __init__(self):
        self.functions = {}
        self.filters = {}


library = _Library()


def jinja2_filter(name=None):
    def closure(func):
        nonlocal name
        if name is None:
            name = func.__name__
        library.filters[name] = func
        return func
    return closure


def jinja2_function(template=None, name=None, takes_context=False, mark_safe=False):
    def closure(orig_func):
        func = orig_func
        wrapper = None
        name_ = name or getattr(func,'_decorated_function', func).__name__

        if template:
            def wrapper(*args, **kwargs):
                from django.template.loader import get_template
                t = get_template(template)
                ctx = orig_func(*args, **kwargs)
                result = t.render(ctx)
                result = jinja2.Markup(result)
                return result
        elif mark_safe:
            def wrapper(*args, **kwargs):
                result = orig_func(*args, **kwargs)
                return jinja2.Markup(result)

        if wrapper:
            func = functools.update_wrapper(wrapper, func)

        if takes_context:
            func = jinja2.contextfunction(func)

        library.functions[name_] = func
        return orig_func

    return closure


from collections import OrderedDict

from django.core.exceptions import FieldDoesNotExist
from django.utils import six
from django.db import models
from .formatters import *
from .base import FormattedObject, FormattedTable
from django.conf import settings


__all__ = ['FormattedModel', 'FormattedModelTable', 'object_formatter_factory', 'table_formatter_factory']


FORMATTER_MAPPING = {
    models.DateTimeField: DateTimeFormatter,
    models.DateField: DateFormatter,
    models.TimeField: TimeFormatter,
    models.ImageField: ImageFormatter,
    models.FileField: FileFormatter,
    models.FloatField: DecimalFormatter,
    models.DecimalField: DecimalFormatter
}


FORMATTER_MAPPING.update(getattr(settings, 'FIELD_FORMATTERS', {}))


def get_formatter_for_field(meta, field_map, field_name, options=None):
    parts = field_name.split("__")
    field_name = parts.pop(0)
    field = field_map.get(field_name)
    if not field:
        return Formatter()
    while parts and field.model:
        model = field.related_model
        item = parts.pop(0)
        field = model._meta.get_field(item)
    try:
        return getattr(field, 'get_formatter')()
    except AttributeError:
        label = field.verbose_name.title()
        label = getattr(options, 'labels', {}).get(field.name, label)
        if field.choices:
            result = ChoiceFormatter(label=label)
        else:
            result = FORMATTER_MAPPING.get(field.__class__, Formatter)(label=label)
        return result


def get_formatters_for_model(model_class, fields=None, exclude=None, options=None):
    meta = model_class._meta
    field_map = OrderedDict((f.name, f) for f in meta.get_fields() if f.concrete)
    if fields is None:
        fields = list(field_map.keys())
    if exclude:
        exclude = set(exclude)
        fields = filter(lambda f: f not in exclude, fields)
    result = [(f, get_formatter_for_field(meta, field_map, f, options)) for f in fields]
    return result


class MetaFormattedModel(type):

    def __init__(cls, name, bases, attrs):
        super(MetaFormattedModel, cls).__init__(name, bases, attrs)
        meta = getattr(cls, 'Meta', None)
        if not meta:
            return
        model = meta.model

        for name, formatter in get_formatters_for_model(model, fields=getattr(meta, 'fields', None),
                                                        exclude=getattr(meta, 'exclude', None),
                                                        options=meta):
            previous = getattr(cls, name, None)
            if not isinstance(previous, Formatter):
                formatter._update_position()
                setattr(cls, name, formatter)
            else:
                previous._update_position()


@six.add_metaclass(MetaFormattedModel)
class FormattedModel(FormattedObject):
    pass


@six.add_metaclass(MetaFormattedModel)
class FormattedModelTable(FormattedTable):
    pass


def object_formatter_factory(model_class, fields=None, exclude=None, **kwargs):
    name = "Formatted%sObject" % model_class.__name__
    meta = type("Meta", (), {"fields": fields, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (FormattedModel, ), kwargs)


def table_formatter_factory(model_class, fields=None, exclude=None, **kwargs):
    name = "Formatted%sTable" % model_class.__name__
    meta = type("Meta", (), {"fields": fields, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (FormattedModelTable, ), kwargs)


from collections import OrderedDict

import re
from django import forms
from django.utils import six
from django.utils.encoding import force_text
from .base import Formatter

__all__ = ['SortFilterField']


class SortFilterField(forms.ChoiceField):

    def __init__(self, choices=(), toggle=True, **kwargs):
        kwargs.setdefault("required", False)
        kwargs.setdefault("widget", forms.HiddenInput)
        super(SortFilterField, self).__init__(choices=choices, **kwargs)
        self.toggle = toggle
        self.set_choices(choices)

    def set_choices(self, choices):
        field_map = {}
        new_choices = []
        for i, (value, label) in enumerate(choices):
            position = str(i)
            new_choices.append((position, label))
            field_map[position] = re.sub(r'\s+', ' ', value.strip())
        self.choices = tuple(new_choices)
        self.field_map = field_map

    def valid_value(self, value):
        "Check to see if the provided value is a valid choice"
        text_value = force_text(value)
        if text_value.startswith("-"):
            text_value = text_value[1:]
        return text_value in self.field_map or super(SortFilterField, self).valid_value(text_value)

    def invert_sign(self, name, neg):
        if name.startswith("-"):
            neg = not neg
        return "%s%s" % ("-" if neg else "", name.lstrip("-"))

    def bind_form(self, form):
        formatter_class = form.context.get('formatter_class')
        if formatter_class is None:
            return
        column_dict = OrderedDict(Formatter.extract_from(formatter_class))
        choices = [(name, col.label or name.title()) for name, col in column_dict.items()
                        if col.sortable and not col.hidden]
        self.column_dict = column_dict
        self.set_choices(choices)

    def filter_queryset(self, queryset, key, bound_field):
        neg = key.startswith("-")
        value = self.field_map[key.lstrip("-")]
        invert = lambda a: self.invert_sign(a, neg)
        values = map(invert, value.split())
        return queryset.order_by(*values)

    def filter_queryset(self, queryset, value, bound_field):
        text_value = force_text(value) if value is not None else None
        if not text_value:
            return queryset
        reverse = text_value.startswith("-")
        column_dict = self.column_dict
        name = text_value[1:] if reverse else text_value
        name = self.field_map[name]
        col = column_dict[name]
        if not col.sortable:
            return queryset
        attr = col.attr or name
        if col.reverse:
            reverse = not reverse
        if reverse:
            attr = "-%s" % attr
        return queryset.order_by(attr)


    def get_value_for_name(self, name):
        for value, key in six.iteritems(self.field_map):
            if name == key:
                return value